# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2021 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import logging as log
from argparse import ArgumentParser, HelpFormatter
from .config import GlobalConfig
from .utils import print_error
from .utils.env import set_unicode_allowed, set_owning_user
from .osbase import OSBase


def init_config(options):
    '''
    Create a new :GlobalConfig from command-line options.
    '''
    gconf = GlobalConfig()
    gconf.load(options.config)

    # check if we are forbidden from using unicode - otherwise we build
    # with unicode enabled by default
    if options.no_unicode:
        set_unicode_allowed(False)
    else:
        import locale
        current_encoding = locale.getpreferredencoding()
        if current_encoding.lower() != 'utf-8':
            log.warning(('Building with unicode support, but your environment does not seem to support unicode. '
                         '(Encoding is {})').format(current_encoding))
        set_unicode_allowed(True)

    if options.owner:
        info = options.owner.split(':')
        if len(info) > 2:
            print_error('You can only use one colon to split user:group when using the --owner flag.')
            sys.exit(1)
        if len(info) == 2:
            user = info[0]
            group = info[1]
        else:
            user = info[0]
            group = None
        set_owning_user(user, group)

    return gconf


def check_print_version(options):
    if options.show_version:
        from . import __version__
        print(__version__)
        sys.exit(0)


def command_create(options):
    ''' Create new container image '''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container name (suite name) to bootstrap!')
        sys.exit(1)
    gconf = init_config(options)

    components = None
    if options.components:
        components = options.components.split(',')

    extra_suites = []
    if options.extra_suites:
        extra_suites = options.extra_suites.strip().split(' ')

    osbase = OSBase(gconf,
                    options.suite,
                    options.arch,
                    variant=options.variant,
                    base_suite=options.base_suite,
                    custom_name=options.name)
    r = osbase.create(options.mirror,
                      components,
                      extra_suites=extra_suites,
                      extra_source_lines=options.extra_source_lines,
                      allow_recommends=options.allow_recommends)
    if not r:
        sys.exit(2)


def command_delete(options):
    ''' Delete container image '''

    check_print_version(options)
    if not options.name:
        print_error('No container image name was specified!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf,
                    options.suite,
                    options.arch,
                    options.variant,
                    custom_name=options.name)
    r = osbase.delete()
    if not r:
        sys.exit(2)


def command_update(options):
    ''' Update container image '''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image name for update!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf,
                    options.suite,
                    options.arch,
                    options.variant,
                    custom_name=options.name)
    if options.recreate:
        r = osbase.recreate()
    else:
        r = osbase.update()
    if not r:
        sys.exit(2)


def command_list(options):
    ''' List container images '''

    from .osbase import print_container_base_image_info

    check_print_version(options)
    gconf = init_config(options)
    print_container_base_image_info(gconf)


def command_build(options):
    ''' Build a package in a new volatile container '''

    from .build import build_from_directory, build_from_dsc

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image or suite name for building!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf,
                    options.suite,
                    options.arch,
                    options.variant,
                    custom_name=options.name)

    # prepare user-defined environment variables
    env_vars = {}
    if options.env_vars:
        for kv in options.env_vars:
            p = kv.split('=', 1)
            if len(p) != 2:
                print('Environment variable definition `{}` is invalid!'.format(kv))
                print('Can not continue.')
                sys.exit(1)
            env_vars[p[0]] = p[1]

    # prepare user-defined buildflags
    buildflags = []
    if options.buildflags:
        buildflags = options.buildflags.split(';')

    if not options.target and os.path.isdir(options.name):
        print('A directory is given as parameter, but you are missing a container-name parameter to build for.')
        print('Can not continue.')
        sys.exit(1)

    # override globally configured output directory with
    # a custom one defined on the CLI
    if options.results_dir:
        osbase.results_dir = options.results_dir

    if not options.target or os.path.isdir(options.target):
        r = build_from_directory(osbase,
                                 options.target,
                                 sign=options.sign,
                                 build_only=options.build_only,
                                 include_orig=options.include_orig,
                                 maintainer=options.maintainer,
                                 clean_source=options.clean_source,
                                 qa_lintian=options.lintian,
                                 interact=options.interact,
                                 log_build=not options.no_buildlog,
                                 extra_dpkg_flags=buildflags,
                                 build_env=env_vars)
    else:
        r = build_from_dsc(osbase,
                           options.target,
                           sign=options.sign,
                           build_only=options.build_only,
                           include_orig=options.include_orig,
                           maintainer=options.maintainer,
                           qa_lintian=options.lintian,
                           interact=options.interact,
                           log_build=not options.no_buildlog,
                           extra_dpkg_flags=buildflags,
                           build_env=env_vars)
    if not r:
        sys.exit(2)


def command_login(options):
    ''' Open interactive session in a container '''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image name!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf,
                    options.suite,
                    options.arch,
                    options.variant,
                    custom_name=options.name)

    allowed = []
    if options.allow:
        allowed = [s.strip() for s in options.allow.split(',')]
    r = osbase.login(options.persistent, allowed)
    if not r:
        sys.exit(2)


def command_run(options, custom_command):
    ''' Run arbitrary command in container session '''

    check_print_version(options)
    if not options.name:
        print_error('Need at least a container image name!')
        sys.exit(1)
    gconf = init_config(options)
    osbase = OSBase(gconf,
                    options.suite,
                    options.arch,
                    options.variant,
                    custom_name=options.name,
                    cachekey=options.cachekey)

    allowed = []
    if options.allow:
        allowed = [s.strip() for s in options.allow.split(',')]

    bind_build_dir = options.bind_build_dir
    if bind_build_dir == 'y':
        bind_build_dir = 'ro'
    elif bind_build_dir == 'rw' or bind_build_dir == 'ro':
        pass
    else:
        bind_build_dir = 'n'

    r = osbase.run(custom_command,
                   options.build_dir,
                   options.artifacts_dir,
                   init_command=options.init_command,
                   copy_command=options.external_commad,
                   header_msg=options.header,
                   bind_build_dir=bind_build_dir,
                   allowed=allowed)
    if not r:
        sys.exit(2)


def command_maintain(options):
    ''' Execute global maintenance actions '''

    check_print_version(options)
    gconf = init_config(options)

    if options.migrate:
        from .maintain import maintain_migrate
        maintain_migrate(gconf)
        return
    if options.clear_caches:
        from .maintain import maintain_clear_caches
        maintain_clear_caches(gconf)
        return
    if options.update_all:
        from .maintain import maintain_update_all
        maintain_update_all(gconf)
        return
    if options.purge:
        from .maintain import maintain_purge
        maintain_purge(gconf, options.yes)
        return
    if options.status:
        from .maintain import maintain_print_status
        maintain_print_status(gconf)
        return

    print_error('No maintenance action selected!')
    sys.exit(1)


class CustomArgparseFormatter(HelpFormatter):

    def _split_lines(self, text, width):
        if text.startswith('CF|'):
            return text[3:].splitlines()
        return HelpFormatter._split_lines(self, text, width)


def add_container_select_arguments(parser):
    parser.add_argument('--variant', action='store', dest='variant', default=None,
                        help='Set the bootstrap script variant (use `none` to select no variant).')
    parser.add_argument('-a', '--arch', action='store', dest='arch', default=None,
                        help='The architecture of the container.')
    parser.add_argument('--suite', action='store', dest='suite', default=None,
                        help='Explicitly set a suite name (instead of having it derived from the container name).')
    parser.add_argument('name', action='store', nargs='?', default=None,
                        help='The name of the container image (usually a distribution suite name).')


def create_parser(formatter_class=None):
    ''' Create debspawn CLI argument parser '''

    if not formatter_class:
        formatter_class = CustomArgparseFormatter

    parser = ArgumentParser(description='Build in nspawn containers', formatter_class=formatter_class)
    subparsers = parser.add_subparsers(dest='sp_name', title='subcommands')

    # generic arguments
    parser.add_argument('-c', '--config', action='store', dest='config', default=None,
                        help='Path to the global config file.')
    parser.add_argument('--verbose', action='store_true', dest='verbose',
                        help='Enable debug messages.')
    parser.add_argument('--no-unicode', action='store_true', dest='no_unicode',
                        help='Disable unicode support.')
    parser.add_argument('--version', action='store_true', dest='show_version',
                        help='Display the version of debspawn itself.')

    parser.add_argument('--owner', action='store', dest='owner', default=None,
                        help=('Set the user name/uid and group/gid separated by a colon '
                              'whose behalf we are acting.'))

    # 'create' command
    sp = subparsers.add_parser('create', help='Create new container image')
    add_container_select_arguments(sp)
    sp.add_argument('--mirror', action='store', dest='mirror', default=None,
                    help='Set a specific mirror to bootstrap from.')
    sp.add_argument('--components', action='store', dest='components', default=None,
                    help='A comma-separated list of archive components to enable in the newly created image.')
    sp.add_argument('--base-suite', action='store', dest='base_suite', default=None,
                    help=('A full suite that forms the base of the selected partial suite '
                          '(e.g. for -updates and -backports).'))
    sp.add_argument('--extra-suites', action='store', dest='extra_suites', default=None,
                    help=('Space-separated list of additional suites that should also be added to the '
                          'sources.list file.'))
    sp.add_argument('--extra-sourceslist-lines', action='store', dest='extra_source_lines', default=None,
                    help=('Lines that should be added to the build environments source.list verbatim. '
                          'Separate lines by linebreaks.'))
    sp.add_argument('--allow-recommends', action='store', dest='allow_recommends', default=None,
                    help=('Do not disable APT installing "recommends"-type dependencies by default, and instead use '
                          'the default behavior for full, normal system installations with "recommends" enabled.'))
    sp.set_defaults(func=command_create)

    # 'delete' command
    sp = subparsers.add_parser('delete', help='Remove a container image')
    add_container_select_arguments(sp)
    sp.set_defaults(func=command_delete)

    # 'update' command
    sp = subparsers.add_parser('update', help='Update a container image')
    add_container_select_arguments(sp)
    sp.add_argument('--recreate', action='store_true', dest='recreate',
                    help=('Re-create the container image from scratch using the settings used to create it previously, '
                          'instead of just updating it.'))
    sp.set_defaults(func=command_update)

    # 'list' command
    sp = subparsers.add_parser('list', help='List available container images', aliases=['ls'])
    sp.set_defaults(func=command_list)

    # 'build' command
    sp = subparsers.add_parser('build', help='Build a package in an isolated environment',
                               formatter_class=formatter_class, aliases=['b'])
    add_container_select_arguments(sp)
    sp.add_argument('--sign', action='store_true', dest='sign',
                    help='Sign the resulting package.')
    sp.add_argument('--only', choices=['binary', 'arch', 'indep', 'source'], dest='build_only',
                    help=('CF|Select only a specific set of packages to be built. Choices are:\n'
                          'binary: Build only binary packages, no source files are to be built and/or distributed.\n'
                          'arch: Build only architecture-specific binary packages.\n'
                          'indep: Build only architecture-independent (arch:all) binary packages.\n'
                          'source: Do a source-only build, no binary packages are made.'))
    sp.add_argument('--include-orig', action='store_true', dest='include_orig',
                    help='Forces the inclusion of the original source.')
    sp.add_argument('--buildflags', action='store', dest='buildflags',
                    help='Set flags passed through to dpkg-buildpackage as semicolon-separated list.')
    sp.add_argument('--results-dir', action='store', dest='results_dir',
                    help='Override the configured results directory and return artifacts at a custom location.')
    sp.add_argument('--maintainer', action='store', dest='maintainer',
                    help=('Set the name and email address of the maintainer for this package and upload, rather than '
                          'using the information from the source tree\'s control file or changelog.'))
    sp.add_argument('--clean-source', action='store_true', dest='clean_source',
                    help=('Run the d/rules clean target outside of the container. This means the package build '
                          'dependencies need to be installed on the host system when building from a '
                          'source directory.'))
    sp.add_argument('--lintian', action='store_true', dest='lintian',
                    help='Run the Lintian static analysis tool for Debian packages after the package is built.')
    sp.add_argument('--no-buildlog', action='store_true', dest='no_buildlog',
                    help='Do not write a build log.')
    sp.add_argument('-i', '--interact', action='store_true', dest='interact',
                    help=('Run an interactive shell in the build environment after build. This implies `--no-buildlog` '
                          'and disables the log.'))
    sp.add_argument('-e', '--setenv', action='append', dest='env_vars',
                    help=('Set an environment variable for the build process. Takes a `key=value` pair any may be '
                          'defined multiple times to set different environment variables.'))
    sp.add_argument('target', action='store', nargs='?', default=None,
                    help='The source package file or source directory to build.')
    sp.set_defaults(func=command_build)

    # 'login' command
    sp = subparsers.add_parser('login', help='Open interactive session in a container')
    add_container_select_arguments(sp)
    sp.add_argument('--persistent', action='store_true', dest='persistent',
                    help='Make changes done in the session persistent.')
    sp.add_argument('--allow', action='store', dest='allow',
                    help=('List one or more additional permissions to grant the container. Takes a comma-separated '
                          'list of capability names.'))
    sp.set_defaults(func=command_login)

    # 'run' command
    sp = subparsers.add_parser('run', help='Run arbitrary command in an ephemeral container')
    add_container_select_arguments(sp)
    sp.add_argument('--artifacts-out', action='store', dest='artifacts_dir', default=None,
                    help='Directory on the host where artifacts can be stored. Mounted to /srv/artifacts in the guest.')
    sp.add_argument('--build-dir', action='store', dest='build_dir', default=None,
                    help='Select a host directory that gets copied to /srv/build.')
    sp.add_argument('--bind-build-dir', action='store', dest='bind_build_dir', default='n',
                    choices=['y', 'n', 'ro', 'rw'],
                    help=('Bindmount build directory instead of copying it. Mounts read-only by default, but can mount '
                          'as writable as well if \'rw\' is passed as value.'))
    sp.add_argument('--cachekey', action='store', dest='cachekey', default=None,
                    help=('If set, use the specified cache-ID to store an initialized container image for faster '
                          'initialization times.\n'
                          'This may mean that the command passed in `--init-command` is skipped if the cache '
                          'already existed.'))
    sp.add_argument('--init-command', action='store', dest='init_command', default=None,
                    help='The command or command script used to set up the container.')
    sp.add_argument('-x', '--external-command', action='store_true', dest='external_commad',
                    help=('If set, the command script(s) will be copied from the host to the container '
                          'and then executed.'))
    sp.add_argument('--header', action='store', dest='header', default=None,
                    help='Name of the task that is run, will be printed as header.')
    sp.add_argument('--allow', action='store', dest='allow',
                    help=('List one or more additional permissions to grant the container. Takes a comma-separated '
                          'list of capability names.'))
    sp.add_argument('command', action='store', nargs='*', default=None,
                    help='The command to run.')

    # 'maintain' command
    sp = subparsers.add_parser('maintain', help='Execute various maintenance actions, affecting all images')
    sp.add_argument('-y', '--yes', action='store_true', dest='yes',
                    help='Perform dangerous actions without asking twice.')
    sp.add_argument('--migrate', action='store_true', dest='migrate',
                    help='Migrate any settings or configuration changes to the current version of debspawn.')
    sp.add_argument('--update-all', action='store_true', dest='update_all',
                    help='Update all container images that we know.')
    sp.add_argument('--clear-caches', action='store_true', dest='clear_caches',
                    help='Delete all cached packages for all images.')
    sp.add_argument('--purge', action='store_true', dest='purge',
                    help='Remove all images as well as any data associated with them.')
    sp.add_argument('--status', action='store_true', dest='status',
                    help='Display a status summary about this installation, highlighting potential issues.')
    sp.set_defaults(func=command_maintain)

    return parser


def run(mainfile, args):
    if len(args) == 0:
        print_error('Need a subcommand to proceed!')
        sys.exit(1)

    parser = create_parser()

    # special case, so 'run' can understand which arguments are for debspawn and which are
    # for the command to be executed
    custom_command = None
    if args[0] == 'run':
        for i, arg in enumerate(args):
            if arg == '---':
                if i + 1 == len(args):
                    print_error('No command was given after "---", can not continue.')
                    sys.exit(1)
                custom_command = args[i + 1:]
                args = args[:i]
                break

    args = parser.parse_args(args)
    check_print_version(args)
    if args.sp_name == 'run':
        if not custom_command:
            custom_command = args.command
        command_run(args, custom_command)
    else:
        if not hasattr(args, 'func'):
            print_error('Unknown or no subcommand was provided. Can not proceed.')
            sys.exit(1)
        args.func(args)



# -*- coding: utf-8 -*-
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

from .log import print_info, print_warn, print_error, print_header, print_section
from .env import colored_output_allowed, unicode_allowed
from .command import safe_run, run_forwarded
from .misc import listify, temp_dir, cd, hardlink_or_copy, format_filesize, rmtree_mntsafe

__all__ = ['print_info',
           'print_warn',
           'print_error',
           'print_header',
           'print_section',
           'colored_output_allowed',
           'unicode_allowed',
           'safe_run',
           'run_forwarded',
           'listify',
           'temp_dir',
           'cd',
           'hardlink_or_copy',
           'format_filesize',
           'rmtree_mntsafe']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2021 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import subprocess
import shutil
from pathlib import Path
from contextlib import contextmanager
from typing import Optional
from .utils import temp_dir, print_header, print_section, format_filesize, \
    print_info, print_error, print_warn, listify
from .utils.misc import maybe_remove, safe_copy
from .utils.env import ensure_root, get_random_free_uid_gid, get_owner_uid_gid
from .utils.command import safe_run
from .utils.zstd_tar import compress_directory, decompress_tarball, ensure_tar_zstd
from .nspawn import nspawn_run_helper_persist, nspawn_run_persist
from .aptcache import APTCache


def debootstrap_version():
    ds_version = 'unknown'
    try:
        out, _, _ = safe_run(['debootstrap', '--version'])
        parts = out.strip().split(' ', 2)
        ds_version = parts[0 if len(parts) < 2 else 1]
    except Exception as e:
        print_warn('Unable to determine debootstrap version: {}'.format(e))

    return ds_version


class OSBase:
    '''
    Describes an OS base registered with debspawn
    '''

    def __init__(self, gconf, suite, arch, variant=None, *,
                 base_suite=None, custom_name=None, cachekey=None):
        self._gconf = gconf
        self._suite = suite
        self._base_suite = base_suite
        self._arch = arch

        # if variant is not passed, we use the configured default
        self._variant = variant if variant else gconf.default_bootstrap_variant
        if self._variant == 'none':
            # "none" is an alias to "don't set a variant when invoking debootstrap"
            self._variant = None

        self._custom_name = custom_name
        self._name = self._make_name()
        self._results_dir = self._gconf.results_dir
        self._cachekey = cachekey
        if self._cachekey:
            self._cachekey = self._cachekey.replace(' ', '')

        self._parameters_checked = False
        self._aptcache = APTCache(self)

        # get a fresh UID to give to our build user within the container
        self._builder_uid = get_random_free_uid_gid()[0]

        # ensure we can (de)compress zstd tarballs
        ensure_tar_zstd()

    def _make_name(self):
        ''' Configure a unique-ish name based on user defined data,
        and tweak the custom name and suite values to match. '''

        if self._custom_name and not self._suite:
            # if we have a custom name but no suite name, the custom name is treated
            # as our suite name *if* no image exists with the custom name
            # (this is for backwards compatibility)
            self._name = self._custom_name
            if not self.exists():
                self._suite = self._custom_name
                self._custom_name = None
        if self._custom_name == self._suite:
            self._custom_name = None

        if not self._arch:
            out, _, ret = safe_run(['dpkg', '--print-architecture'])
            if ret != 0:
                raise Exception('Running dpkg --print-architecture failed: {}'.format(out))

            self._arch = out.strip()
        if self._custom_name:
            return self._custom_name
        elif self._variant:
            return '{}-{}-{}'.format(self._suite, self._variant, self._arch)
        else:
            return '{}-{}'.format(self._suite, self._arch)

    def _custom_name_parameter_check(self):
        ''' Read parameters in case a custom name was passed, and perform basic sanity checks. '''
        import sys

        if self._parameters_checked:
            return
        if not self._custom_name:
            return

        config_fname = self.get_config_location()
        if not os.path.isfile(config_fname):
            print_error('No configuration data found for image "{}"!'.format(self.name))
            sys.exit(3)

        with open(config_fname, 'rt') as f:
            cdata = json.loads(f.read())

            c_suite = cdata.get('Suite', self.suite)
            if not self._suite:
                # if no suite was set, but we have one in the manifest file,
                # we will always use it to fill in the gap
                self._suite = c_suite

            if self.suite != c_suite:
                print_error('Expected suite name "{}" for image "{}", but got "{}" instead.'.format(
                    cdata.get('Suite'), self.name, self.suite))
                sys.exit(1)
            c_arch = cdata.get('Architecture', self.arch)
            c_variant = cdata.get('Variant', self.variant)

            if self.arch and self.arch != c_arch:
                print_warn(('Expected architecture "{}" for image "{}", but got "{}" instead. '
                            'Using expected value.').format(
                                c_arch, self.name, self.arch))
            if self.variant and self.variant != c_variant:
                print_warn(('Expected variant "{}" for image "{}", but got "{}" instead. '
                            'Using expected value.').format(
                                c_variant, self.name, self.variant))
            self._arch = c_arch
            self._variant = c_variant
        self._parameters_checked = True

    @property
    def name(self) -> str:
        return self._name

    @property
    def suite(self) -> str:
        return self._suite

    @property
    def base_suite(self) -> str:
        return self._base_suite

    @property
    def arch(self) -> str:
        return self._arch

    @property
    def variant(self) -> str:
        return self._variant

    @property
    def global_config(self):
        return self._gconf

    @property
    def aptcache(self):
        return self._aptcache

    @property
    def has_base_suite(self) -> bool:
        return True if self.base_suite and self.base_suite != self.suite else False

    @property
    def results_dir(self):
        Path(self._results_dir).mkdir(parents=True, exist_ok=True)
        return self._results_dir

    @results_dir.setter
    def results_dir(self, path):
        self._results_dir = path
        Path(self._results_dir).mkdir(exist_ok=True)

    def _copy_helper_script(self, osroot_path):
        script_location = os.path.join(osroot_path, 'usr', 'lib', 'debspawn')
        Path(script_location).mkdir(parents=True, exist_ok=True)
        script_fname = os.path.join(script_location, 'dsrun')

        if os.path.isfile(script_fname):
            os.remove(script_fname)
        shutil.copy2(self._gconf.dsrun_path, script_fname)

        os.chmod(script_fname, 0o0755)

    def get_image_location(self):
        return os.path.join(self._gconf.osroots_dir, '{}.tar.zst'.format(self.name))

    def get_image_cache_dir(self):
        cache_img_dir = os.path.join(self._gconf.osroots_dir, 'dcache', self.name)
        Path(cache_img_dir).mkdir(parents=True, exist_ok=True)
        return cache_img_dir

    def get_cache_image_location(self):
        if not self._cachekey:
            return None
        return os.path.join(self.get_image_cache_dir(), '{}.tar.zst'.format(self._cachekey))

    def get_config_location(self):
        return os.path.join(self._gconf.osroots_dir, '{}.json'.format(self.name))

    def exists(self):
        return os.path.isfile(self.get_image_location())

    def cacheimg_exists(self):
        location = self.get_cache_image_location()
        if not location:
            return False
        return os.path.isfile(location)

    def ensure_exists(self):
        '''
        Ensure the container image exists, and terminate the
        program with an error code in case it does not.
        '''
        import sys
        if not self._load_existent():
            print_error('The container image for "{}" does not exist. Please create it first.'.format(self.name))
            sys.exit(3)

    def _load_existent(self) -> bool:
        ''' Check if image exists, and if so load some essential data and return True. '''
        # ensure the set config values are sane if the user is using a custom container name
        if self._custom_name:
            self._custom_name_parameter_check()
        return self.exists()

    def new_nspawn_machine_name(self):
        import platform
        from random import choice
        from string import ascii_lowercase, digits

        nid = ''.join(choice(ascii_lowercase + digits) for _ in range(4))

        # on Linux, the maximum hostname length is 64, so we simple set this as general default for
        # debspawn here.
        # shorten the hostname part or replace the suffix, depending on what is longer.
        # This should only ever matter if the hostname of the system already is incredibly long
        uniq_suffix = '{}-{}'.format(self.name, nid)
        if len(uniq_suffix) > 48:
            uniq_suffix = ''.join(choice(ascii_lowercase + digits) for _ in range(12))
        node_name_prefix = platform.node()[:63 - len(uniq_suffix)]

        return '{}-{}'.format(node_name_prefix, uniq_suffix)

    def _write_config_json(self, mirror, components, extra_suites, extra_source_lines):
        '''
        Create configuration file for this container base image
        '''

        print_info('Saving configuration settings.')
        data = {'Suite': self.suite,
                'Architecture': self.arch}
        if self.variant:
            data['Variant'] = self.variant
        if mirror:
            data['Mirror'] = mirror
        if components:
            data['Components'] = components
        if extra_suites:
            data['ExtraSuites'] = extra_suites
        if extra_source_lines:
            data['ExtraSourceLines'] = extra_source_lines

        with open(self.get_config_location(), 'wt') as f:
            f.write(json.dumps(data, sort_keys=True, indent=4))
            f.write('\n')

    def _clear_image_tree(self, image_dir):
        ''' Clear files from a directory tree that we don't want in the tarball. '''

        if os.path.ismount(image_dir):
            print_warn('Preparing OS tree for compression, but /dev is still mounted.')
            return

        for sdir, _, files in os.walk(os.path.join(image_dir, 'dev')):
            for f in files:
                fname = os.path.join(sdir, f)
                if os.path.lexists(fname) and not os.path.isdir(fname) and not os.path.ismount(fname):
                    os.remove(fname)

    def _remove_unwanted_files(self, instance_dir):
        ''' Delete unwanted files from a base image '''

        # drop resolv.conf: Some OSes set this to a symlink, which will lead to nowhere
        # in the container and may cause issues when bindmounting over it
        maybe_remove(os.path.join(instance_dir, 'etc', 'resolv.conf'))

        # our APT proxy connfiguration should also not be stored, we will reset it every time
        maybe_remove(os.path.join(instance_dir, 'etc', 'apt', 'apt.conf.d', '98debspawn_proxy'))

        # drop machine-id file, if one exists
        maybe_remove(os.path.join(instance_dir, 'etc', 'machine-id'))
        # drop logfiles which we want to reset
        maybe_remove(os.path.join(instance_dir, 'var', 'log', 'lastlog'))
        maybe_remove(os.path.join(instance_dir, 'var', 'log', 'faillog'))

    def _setup_apt_proxy(self, instance_dir):
        ''' Setup APT proxy configuration.
        APT needs special treatment even in the container to work with a company proxy (yuck!),
        so we do this here so the user does not need to care about it.
        '''

        http_proxy = os.getenv('HTTP_PROXY', os.getenv('http_proxy'))
        https_proxy = os.getenv('HTTPS_PROXY', os.getenv('https_proxy'))
        if not http_proxy and not https_proxy:
            return
        if http_proxy:
            http_proxy = http_proxy.replace('"', '')
        if https_proxy:
            https_proxy = https_proxy.replace('"', '')

        proxyconf_fname = os.path.join(instance_dir, 'etc', 'apt', 'apt.conf.d', '98debspawn_proxy')
        with open(proxyconf_fname, 'w') as f:
            if http_proxy:
                f.write('Acquire::http::Proxy "{}";\n'.format(http_proxy))
            if https_proxy:
                f.write('Acquire::https::Proxy "{}";\n'.format(https_proxy))

    def _create_internal(self, mirror=None, components=None,
                         extra_suites: list[str] = None, extra_source_lines: str = None,
                         allow_recommends: bool = False,
                         show_header: bool = True):
        ''' Create new container base image (internal method) '''

        if self.exists():
            print_error('An image already exists for this configuration. Can not create a new one.')
            return False
        if not extra_suites:
            extra_suites = []

        # ensure image location exists
        Path(self._gconf.osroots_dir).mkdir(parents=True, exist_ok=True)

        if show_header:
            print_header('Creating new base: {} [{}]'.format(self.suite, self.arch))
        else:
            print_section('Creating new base: {} [{}]'.format(self.suite, self.arch))

        if self._custom_name:
            print('Custom name: {}'.format(self.name))
        print('Using mirror: {}'.format(mirror if mirror else 'default'))
        if self.variant:
            print('variant: {}'.format(self.variant))
        cmd = ['debootstrap',
               '--arch={}'.format(self.arch),
               '--include=python3-minimal,eatmydata']
        if components:
            cmd.append('--components={}'.format(','.join(components)))
        if self.variant:
            cmd.append('--variant={}'.format(self.variant))

        with temp_dir() as tdir:
            bootstrap_suite = self.suite
            if self.has_base_suite:
                bootstrap_suite = self.base_suite
            cmd.extend([bootstrap_suite, tdir])
            print('Bootstrap suite: {}'.format(bootstrap_suite))
            if extra_suites:
                print('Additional suites: {}'.format(', '.join(extra_suites)))
            if extra_source_lines:
                print('Custom sources.list lines will be added:')
                for line in extra_source_lines.split('\\n'):
                    print('    {}'.format(line))
            if mirror:
                cmd.append(mirror)

            print_section('Bootstrap')
            proc = subprocess.run(cmd, check=False)
            if proc.returncode != 0:
                return False

            # create helper script runner
            self._copy_helper_script(tdir)

            # if we bootstrapped the base suite, add the primary suite to
            # sources.list. We also add any explicit extra suites and source lines
            if self.has_base_suite or extra_suites or extra_source_lines:
                import re

                sourceslist_fname = os.path.join(tdir, 'etc', 'apt', 'sources.list')
                if not mirror:
                    with open(sourceslist_fname, 'r') as f:
                        contents = f.read()
                        matches = re.findall(
                            'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+',
                            contents)
                        if not matches:
                            print_error('Unable to detect default APT repository URL (no regex matches).')
                            return False
                        mirror = matches[0]
                    if not mirror:
                        print_error('Unable to detect default APT repository URL.')
                        return False

                if not components:
                    # FIXME: We should really be more clever here, e.g. depend on python-apt
                    # and parse sources.list properly
                    components = ['main']
                with open(sourceslist_fname, 'a') as f:
                    if self.has_base_suite:
                        f.write('deb {mirror} {suite} {components}\n'.format(mirror=mirror,
                                                                             suite=self.suite,
                                                                             components=' '.join(components)))

                    if extra_suites:
                        f.write('\n')
                        for esuite in extra_suites:
                            if esuite == self.suite or esuite == bootstrap_suite:
                                # don't add existing suites multiple times
                                continue
                            f.write('deb {mirror} {esuite} {components}\n'.format(mirror=mirror,
                                                                                  esuite=esuite,
                                                                                  components=' '.join(components)))

                    if extra_source_lines:
                        f.write('\n')
                        for line in extra_source_lines.split('\\n'):
                            f.write('{}\n'.format(line.strip()))

            # write our default APT settings for this container
            aptconf_fname = os.path.join(tdir, 'etc', 'apt', 'apt.conf.d', '99debspawn')
            with open(aptconf_fname, 'w') as f:
                # fail immediately with a proper exit code when e.g. apt update fails,
                # so we can retry and don't silently use old packages
                # (only available with newer APT versions)
                f.write('APT::Update::Error-Mode "any";\n')
                if not allow_recommends:
                    f.write('APT::Install-Recommends "0";\n')
                    f.write('APT::Install-Suggests "0";\n')

            # delete unwanted files, especially resolv.conf as a broken one will
            # mess with the next step
            self._remove_unwanted_files(tdir)

            # configure APT proxy, so the configure operation will work behind proxys
            self._setup_apt_proxy(tdir)

            print_section('Configure')
            if nspawn_run_helper_persist(self, tdir,
                                         self.new_nspawn_machine_name(),
                                         '--update',
                                         build_uid=self._builder_uid) != 0:
                return False

            # drop any unwanted files (again) before building the tarball
            self._remove_unwanted_files(tdir)

            print_section('Creating Tarball')
            self._clear_image_tree(tdir)
            compress_directory(tdir, self.get_image_location())

        # store configuration settings, so we can later recreate this tarball
        # or just display information about it
        self._write_config_json(mirror, components, extra_suites, extra_source_lines)

        return True

    def create(self, mirror: str = None, components: list[str] = None, *,
               extra_suites: list[str] = None, extra_source_lines: str = None,
               allow_recommends: bool = False):
        ''' Create new container base image (internal method) '''
        ensure_root()

        if self.exists():
            print_error('This configuration has already been created. You can only delete or update it.')
            return False

        ret = self._create_internal(mirror=mirror,
                                    components=components,
                                    extra_suites=extra_suites,
                                    extra_source_lines=extra_source_lines,
                                    allow_recommends=allow_recommends,
                                    show_header=True)
        if ret:
            print_info('Done.')

        return ret

    def delete(self):
        ''' Remove container base image '''
        ensure_root()

        if not self._load_existent():
            print_error('Can not delete "{}": This configuration does not exist.'.format(self.name))
            return False

        print_header('Removing base image {}'.format(self.name))

        print_section('Deleting cache')
        # remove packages cache
        cache_size = self._aptcache.clear()
        print_info('Removed {} cached packages.'.format(cache_size))
        self._aptcache.delete()
        # remove cached images
        shutil.rmtree(self.get_image_cache_dir())
        print_info('Cache directory removed.')

        print_section('Deleting base tarball')
        os.remove(self.get_image_location())

        config_fname = self.get_config_location()
        if os.path.isfile(config_fname):
            print_section('Deleting configuration manifest')
            os.remove(config_fname)

        print_info('Done.')
        return True

    @contextmanager
    def new_instance(self, basename=None):
        with temp_dir() as tdir:
            if self.cacheimg_exists():
                image_fname = self.get_cache_image_location()
            else:
                image_fname = self.get_image_location()

            decompress_tarball(image_fname, tdir)
            self._setup_apt_proxy(tdir)

            yield tdir, self.new_nspawn_machine_name()

    def make_instance_permanent(self, instance_dir):
        ''' Add changes done in the current instance to the main tarball of this OS tree, replacing it. '''

        # remove unwanted files from the tarball
        self._clear_image_tree(instance_dir)

        if self._cachekey:
            tarball_name = self.get_cache_image_location()
        else:
            tarball_name = self.get_image_location()
        tarball_name_old = '{}.old'.format(tarball_name)

        if os.path.isfile(tarball_name):
            os.replace(tarball_name, tarball_name_old)
        compress_directory(instance_dir, tarball_name)
        if os.path.isfile(tarball_name_old):
            os.remove(tarball_name_old)

        tar_size = os.path.getsize(tarball_name)
        if self._cachekey:
            print_info('New compressed tarball size (for {}) is {}'.format(self._cachekey, format_filesize(tar_size)))
        else:
            print_info('New compressed tarball size is {}'.format(format_filesize(tar_size)))

    def update(self):
        ''' Update container base image '''
        ensure_root()

        if not self._load_existent():
            print_error('Can not update "{}": This configuration does not exist.'.format(self.name))
            return False

        print_header('Updating container image')

        with self.new_instance() as (instance_dir, _):
            # ensure helper script runner exists and is up to date
            self._copy_helper_script(instance_dir)

            print_section('Update')
            if nspawn_run_helper_persist(self, instance_dir,
                                         self.new_nspawn_machine_name(),
                                         '--update',
                                         build_uid=self._builder_uid) != 0:
                return False

            # drop unwanted files from the image
            self._remove_unwanted_files(instance_dir)

            print_section('Recreating tarball')
            self.make_instance_permanent(instance_dir)

        print_section('Cleaning up cache')
        cache_size = self._aptcache.clear()
        print_info('Removed {} cached packages.'.format(cache_size))
        # remove now-outdated cached images
        shutil.rmtree(self.get_image_cache_dir())

        print_info('Done.')
        return True

    def recreate(self):
        ''' Recreate a container base image '''
        ensure_root()

        if not self._load_existent():
            print_error('Can not recreate "{}": The image does not exist.'.format(self.name))
            return False

        config_fname = self.get_config_location()
        if not os.path.isfile(config_fname):
            print_error('Can not recreate "{}": Unable to find configuration data for this image.'.format(self.name))
            return False

        print_header('Recreating container image')

        # read configuration data
        with open(config_fname, 'rt') as f:
            cdata = json.loads(f.read())
            self._suite = cdata.get('Suite', self.suite)
            self._arch = cdata.get('Architecture', self.arch)
            self._variant = cdata.get('Variant', self.variant)
            mirror = cdata.get('Mirror')
            components = cdata.get('Components')
            extra_suites = cdata.get('ExtraSuites', [])
            extra_source_lines = cdata.get('ExtraSourceLines')
            allow_recommends = cdata.get('AllowRecommends', False)

        print_section('Deleting cache')
        cache_size = self._aptcache.clear()
        print_info('Removed {} cached packages.'.format(cache_size))
        self._aptcache.delete()
        print_info('Cache directory removed.')

        # move old image tarball out of the way
        image_name = self.get_image_location()
        image_name_old = self.get_image_location() + '.old'
        if os.path.isfile(image_name_old):
            print_info('Removing cruft image')
            os.remove(image_name_old)
        os.rename(image_name, image_name_old)
        print_info('Old tarball moved.')

        # ty to create the tarball again
        try:
            ret = self._create_internal(mirror=mirror,
                                        components=components,
                                        extra_suites=extra_suites,
                                        extra_source_lines=extra_source_lines,
                                        allow_recommends=allow_recommends,
                                        show_header=False)
        except Exception as e:
            print_error('Error while trying to create image: {}'.format(str(e)))
            ret = False

        if ret:
            if os.path.isfile(image_name_old):
                print_info('Removing old image')
                os.remove(image_name_old)

            print_info('Removing outdated cached images')
            shutil.rmtree(self.get_image_cache_dir())

            print_info('Done.')
            return True
        else:
            print_info('Restoring old tarball')
            if os.path.isfile(image_name):
                print_info('Removing failed new image')
                os.remove(image_name)
            os.rename(image_name_old, image_name)
            print_info('Recreation failed.')
            return False

    def login(self, persistent=False, allowed: list[str] = None):
        ''' Interactive shell login into the container '''
        ensure_root()

        if not self._load_existent():
            print_info('Can not enter "{}": This configuration does not exist.'.format(self.name))
            return False

        print_header('Login (persistent changes) for {}'.format(self.name)
                     if persistent else 'Login for {}'.format(self.name))
        with self.new_instance() as (instance_dir, _):
            # ensure helper script runner exists and is up to date
            self._copy_helper_script(instance_dir)

            # run an interactive shell in the new container
            nspawn_run_persist(self,
                               instance_dir,
                               self.new_nspawn_machine_name(),
                               '/srv',
                               verbose=True,
                               allowed=allowed)

            if persistent:
                print_section('Recreating tarball')
                self.make_instance_permanent(instance_dir)
            else:
                print_info('Changes discarded.')

        print_info('Done.')
        return True

    def retrieve_artifacts(self, src_dir: str, dest_dir: Optional[str] = None):
        from glob import glob

        print_section('Retrieving build artifacts')
        if not dest_dir:
            dest_dir = self.results_dir

        o_uid, o_gid = get_owner_uid_gid()
        acount = 0
        for f in glob(os.path.join(src_dir, '*.*')):
            if os.path.isfile(f):
                target_fname = os.path.join(dest_dir, os.path.basename(f))
                safe_copy(f, target_fname)
                os.chown(target_fname, o_uid, o_gid, follow_symlinks=False)
                acount += 1
        print_info('Copied {} files.'.format(acount))

    def _copy_command_script_to_instance_dir(self, instance_dir: str, command_script: str) -> Optional[str]:
        '''
        Copy a script from the host to the current instance directory and make it
        executable.
        Contains the path to the executable script as seen from inside the container.
        '''
        host_script = os.path.abspath(command_script)
        if not os.path.isfile(host_script):
            return None

        script_location = os.path.join(instance_dir, 'srv', 'tmp')
        Path(script_location).mkdir(parents=True, exist_ok=True)
        script_fname = os.path.join(script_location, os.path.basename(host_script))

        if os.path.isfile(script_fname):
            os.remove(script_fname)
        shutil.copy2(host_script, script_fname)
        os.chmod(script_fname, 0o0755)

        return os.path.join('/srv', 'tmp', os.path.basename(host_script))

    def run(self, command, build_dir, artifacts_dir, init_command=None, copy_command=False,
            header_msg=None, bind_build_dir: Optional[str] = None, allowed: list[str] = None):
        ''' Run an arbitrary command or script in the container '''
        ensure_root()

        if not self._load_existent():
            print_error('Can not run command in "{}": The base image does not exist.'.format(self.name))
            return False

        if len(command) <= 0:
            print_error('No command was given. Can not continue.')
            return False

        if isinstance(init_command, str):
            if init_command:
                import shlex
                init_command = shlex.split(init_command)
        init_command = listify(init_command)
        allowed = listify(allowed)
        if bind_build_dir == 'n':
            bind_build_dir = None

        # ensure we have absolute paths
        if build_dir:
            build_dir = os.path.normpath(os.path.abspath(build_dir))
        if artifacts_dir:
            artifacts_dir = os.path.normpath(os.path.abspath(artifacts_dir))

        if self._cachekey and init_command and not self.cacheimg_exists():
            print_header('Preparing template for `{}`'.format(self._cachekey))

            # we do not have a cached image prepared, let's do that now!
            with self.new_instance() as (instance_dir, machine_name):
                # ensure helper script runner exists and is up to date
                self._copy_helper_script(instance_dir)

                if copy_command:
                    # copy initialization script from host to container
                    host_script = init_command[0]
                    init_command[0] = self._copy_command_script_to_instance_dir(instance_dir, host_script)
                    if not init_command[0]:
                        print_error(('Unable to find initialization script "{}", '
                                     'can not copy it to the container. Exiting.').format(host_script))
                        return False

                r = nspawn_run_helper_persist(self,
                                              instance_dir,
                                              machine_name,
                                              '--prepare-run',
                                              '/srv',
                                              build_uid=self._builder_uid)
                if r != 0:
                    print_error('Container setup failed.')
                    return False
                # we do not want some permissions to be in effect here,
                # as they may have unwanted effects on the final cached image
                banned_permissions = ['full-dev', 'full-proc', 'read-kmods']
                filtered_allowed = []
                for perm in allowed:
                    if perm not in banned_permissions:
                        filtered_allowed.append(perm)

                init_nspawn_flags = []
                if build_dir:
                    if not bind_build_dir:
                        shutil.copytree(build_dir, os.path.join(instance_dir, 'srv', 'build'), dirs_exist_ok=True)
                    else:
                        if bind_build_dir == 'rw':
                            init_nspawn_flags = ['--bind={}:/srv/build/'.format(build_dir)]
                        elif bind_build_dir == 'ro':
                            init_nspawn_flags = ['--bind-ro={}:/srv/build/'.format(build_dir)]
                r = nspawn_run_persist(self,
                                       instance_dir,
                                       machine_name,
                                       '/srv',
                                       init_command,
                                       init_nspawn_flags,
                                       allowed=filtered_allowed)
                if r != 0:
                    return False

                print_info('Storing prepared image in cache')
                self.make_instance_permanent(instance_dir)

        if header_msg:
            print_header(header_msg)
        if self._cachekey and init_command and self.cacheimg_exists():
            print_info('Using cached container image `{}`'.format(self._cachekey))

        with self.new_instance() as (instance_dir, machine_name):
            # ensure helper script runner exists and is up to date
            self._copy_helper_script(instance_dir)

            if copy_command:
                # copy the script from the host into our container and execute it there
                host_script = command[0]
                command[0] = self._copy_command_script_to_instance_dir(instance_dir, host_script)
                if not command[0]:
                    print_error(('Unable to find script "{}", can not copy it to the container. '
                                 'Exiting.').format(host_script))
                    return False

            r = nspawn_run_helper_persist(self,
                                          instance_dir,
                                          machine_name,
                                          '--prepare-run',
                                          '/srv',
                                          build_uid=self._builder_uid)
            if r != 0:
                print_error('Container setup failed.')
                return False

            # Create a few directories we may use for bindmounting if some allow-flags are set,
            # and which are not commonly present in the base image.
            # This is only needed for `run` actions, and regular package builds should not require
            # bindmounts to these directories.
            os.makedirs(os.path.join(instance_dir, 'lib', 'modules'), exist_ok=True)
            os.makedirs(os.path.join(instance_dir, 'boot'), exist_ok=True)
            os.makedirs(os.path.join(instance_dir, 'srv', 'artifacts'), exist_ok=True)

            print_section('Running Task')
            nspawn_flags = []
            chdir = '/srv'
            if build_dir:
                chdir = '/srv/build'
                if not bind_build_dir:
                    shutil.copytree(build_dir, os.path.join(instance_dir, 'srv', 'build'), dirs_exist_ok=True)
                else:
                    if bind_build_dir == 'rw':
                        nspawn_flags.extend(['--bind={}:/srv/build/'.format(build_dir)])
                    elif bind_build_dir == 'ro':
                        nspawn_flags.extend(['--bind-ro={}:/srv/build/'.format(build_dir)])

            r = nspawn_run_persist(self,
                                   instance_dir,
                                   machine_name,
                                   chdir,
                                   command,
                                   nspawn_flags,
                                   allowed=allowed)
            if r != 0:
                return False

            # copy results to target directory
            self.retrieve_artifacts(os.path.join(instance_dir, 'srv', 'artifacts'), artifacts_dir)

        print_info('Done.')
        return True


def print_container_base_image_info(gconf):
    '''
    Search for all available container base images and list information
    about them.
    '''
    from glob import glob

    osroots_dir = gconf.osroots_dir
    tar_files = []
    if os.path.isdir(osroots_dir):
        tar_files = list(glob(os.path.join(osroots_dir, '*.tar.zst')))
    if not tar_files:
        print_info('No container base images have been found!')
        return False
    tar_files_len = len(tar_files)

    for i, tar_fname in enumerate(tar_files):
        img_basepath = os.path.splitext(os.path.splitext(tar_fname)[0])[0]
        config_fname = img_basepath + '.json'
        imgid = os.path.basename(img_basepath)
        print('[{}]'.format(imgid))

        cache_files = list(glob(os.path.join(osroots_dir, 'dcache', imgid, '*.tar.zst')))
        cached_names = []
        for cfile in cache_files:
            cname = os.path.basename(os.path.splitext(os.path.splitext(cfile)[0])[0])
            cached_names.append(cname)

        # read configuration data if it exists
        if os.path.isfile(config_fname):
            with open(config_fname, 'rt') as f:
                cdata = json.loads(f.read())
            for key, value in cdata.items():
                if type(value) is list:
                    value = '; '.join(value)
                print('{} = {}'.format(key, value))

        tar_size = os.path.getsize(tar_fname)
        print('Size = {}'.format(format_filesize(tar_size)))

        if cached_names:
            print('CachedImages = {}'.format('; '.join(cached_names)))
        if i != tar_files_len - 1:
            print()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2021 Matthias Klumpp <matthias@tenstral.net>
#
# Licensed under the GNU Lesser General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the license, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import toml


thisfile = __file__
if not os.path.isabs(thisfile):
    thisfile = os.path.normpath(os.path.join(os.getcwd(), thisfile))


__all__ = ['GlobalConfig']


class GlobalConfig:
    '''
    Global configuration singleton affecting all of Debspawn.
    '''

    _instance = None

    class __GlobalConfig:
        def load(self, fname=None):
            if not fname:
                fname = '/etc/debspawn/global.toml'

            cdata = {}
            if os.path.isfile(fname):
                with open(fname) as f:
                    try:
                        cdata = toml.load(f)
                    except toml.TomlDecodeError as e:
                        print('Unable to parse global configuration (global.toml): {}'.format(str(e)), file=sys.stderr)
                        sys.exit(8)

            self._dsrun_path = os.path.normpath(os.path.join(thisfile, '..', 'dsrun'))
            if not os.path.isfile(self._dsrun_path):
                print('Debspawn is not set up properly: Unable to find file "{}". Can not continue.'.format(
                    self._dsrun_path), file=sys.stderr)
                sys.exit(4)

            self._osroots_dir = cdata.get('OSImagesDir', '/var/lib/debspawn/images/')
            self._results_dir = cdata.get('ResultsDir', '/var/lib/debspawn/results/')
            self._aptcache_dir = cdata.get('APTCacheDir', '/var/lib/debspawn/aptcache/')
            self._injected_pkgs_dir = cdata.get('InjectedPkgsDir', '/var/lib/debspawn/injected-pkgs/')
            self._temp_dir = cdata.get('TempDir', '/var/tmp/debspawn/')
            self._default_bootstrap_variant = cdata.get('DefaultBootstrapVariant', 'buildd')
            self._allow_unsafe_perms = cdata.get('AllowUnsafePermissions', False)

            self._syscall_filter = cdata.get('SyscallFilter', 'compat')
            if self._syscall_filter == 'compat':
                # permit some system calls known to be needed by packages that sbuild & Co.
                # build without problems.
                self._syscall_filter = ['@memlock',
                                        '@pkey',
                                        '@clock',
                                        '@cpu-emulation']
            elif self._syscall_filter == 'nspawn-default':
                # make no additional changes, so nspawn's built-in defaults are used
                self._syscall_filter = []
            else:
                if type(self._syscall_filter) is not list:
                    print(('Configuration error (global.toml): Entry "SyscallFilter" needs to be either a string value '
                           '("compat" or "nspawn-default"), or a list of permissible system call names as listed by '
                           'the syscall-filter command of systemd-analyze(1)'), file=sys.stderr)
                    sys.exit(8)

        @property
        def dsrun_path(self) -> str:
            return self._dsrun_path

        @dsrun_path.setter
        def dsrun_path(self, v):
            self._dsrun_path = v

        @property
        def osroots_dir(self) -> str:
            return self._osroots_dir

        @property
        def results_dir(self) -> str:
            return self._results_dir

        @property
        def aptcache_dir(self) -> str:
            return self._aptcache_dir

        @property
        def injected_pkgs_dir(self) -> str:
            return self._injected_pkgs_dir

        @property
        def temp_dir(self) -> str:
            return self._temp_dir

        @property
        def default_bootstrap_variant(self) -> str:
            return self._default_bootstrap_variant

        @property
        def syscall_filter(self) -> list:
            return self._syscall_filter

        @property
        def allow_unsafe_perms(self) -> bool:
            return self._allow_unsafe_perms

    def __init__(self, fname=None):
        if not GlobalConfig._instance:
            GlobalConfig._instance = GlobalConfig.__GlobalConfig()
            GlobalConfig._instance.load(fname)

    def __getattr__(self, name):
        return getattr(self._instance, name)
